(* ICFP 2002 Programming Contest *)
(* Robot Delivery Service *)
(* Grant Glouser <glouser@cs.stanford.edu> *)

(* Basic datatypes of the robot world. *)
type tile = Open | Water | Wall | HomeBase ;;
type direction = North | East | South | West ;;
type command =
    Move of direction
  | Pick of int list
  | Drop of int list;;
type action =
    Moved of direction
  | Picked of int
  | Dropped of int
  | AddRobot of (int * int);;

let tileCharMapping = [ ('.', Open);
                        ('~', Water);
                        ('#', Wall);
                        ('@', HomeBase) ];;
let charTileMapping = List.map (fun (c,t) -> (t,c)) tileCharMapping;;
let tile_of_char c = List.assoc c tileCharMapping;;
let char_of_tile t = List.assoc t charTileMapping;;

let char_of_direction = function
    North -> 'N'
  | South -> 'S'
  | East -> 'E'
  | West -> 'W';;

let next_location (x,y) = function
  | North -> (x,y+1)
  | South -> (x,y-1)
  | East  -> (x+1,y)
  | West  -> (x-1,y);;



module Map =
  struct
    type mapdata = { size : int*int; tiles : tile array array }

    (* For convenience, we wrap the map data with a boundary of Wall tiles. *)
    let wrap_row row width =
      let new_row = Array.make (width+2) Wall in
      Array.blit row 0 new_row 1 width;
      new_row
    let wrap tiles (width,height) = Array.init (height+2) (fun i ->
        if i = 0 || i = height+1 then Array.make (width+2) Wall
        else wrap_row tiles.(i-1) width)

    assert ((wrap [|[|Open|]|] (1,1)) = [| [| Wall;Wall;Wall|];
                                           [| Wall;Open;Wall|];
                                           [| Wall;Wall;Wall|] |])

    (* The map class encapsulates map data. *)
    class map data_init =
      object
        val data = { size = data_init.size;
                     tiles = wrap data_init.tiles data_init.size }
        method data = data
        method get (i,j) = data.tiles.(j).(i)
        method width = fst data.size
        method height = snd data.size
      end

    (* default_map is a basic object of the map class. *)
    let default_map = new map {size=(0,0); tiles=[|[||]|]}
  end;;


module RDSServer =
  struct
    (* Parsing server messages. *)
    let space = parser [< '' ' >] -> ()
    let newline = parser [< ''\n' >] -> ()
    let int_of_digit c = int_of_char c - int_of_char '0'
    let integer =
      let rec integer_acc n = parser
          [< ''0'..'9' as c;
             r = (integer_acc (10*n + int_of_digit c)) >] -> r
        | [< >] -> n
      in parser [< ''0'..'9' as c; s >] -> integer_acc (int_of_digit c) s

    type token = INT of int
      | ID of int
      | ACTCODE of char
      | TILE of tile
      | STOP

    let rec nextActionToken = parser
        [< ''#'; t = (parser
            [< id = integer >] -> Some (ID id)
          | [< >] -> Some (TILE Wall)) >] -> t
      | [< ''N'|'S'|'E'|'W'|'D'|'P'|'X'|'Y' as c >] -> Some (ACTCODE c)
      | [< ''.'|'~'|'@' as c >] -> Some (TILE (tile_of_char c))
      | [< _ = space; s >] -> nextActionToken s
      | [< _ = newline >] -> Some (STOP)
      | [< n = integer >] -> Some (INT n)
      | [< >] -> None

    let tokenize str = Stream.from (fun _ -> nextActionToken str)

    (* Read a row of tiles from the stream. *)
    let rec map_row = parser
        [< 'TILE t; rest = map_row >] -> t :: rest
      | [< 'STOP >] -> []

    (* Read a map with the given dimensions from the stream. *)
    let tiles width height str =
      Array.init height (fun _ -> Array.of_list (map_row str))

    (* Read a whole map from the stream. *)
    let map = parser
        [< 'INT width; 'INT height; 'STOP;
           tiles = (tiles width height) >] ->
             { Map.size = (width,height); Map.tiles = tiles }

    (* Read an initial player configuration from the stream. *)
    let config = parser
        [< 'INT uid; 'INT capacity; 'INT money; 'STOP >] ->
          (uid, capacity, money)

    let action = parser
        [< 'ACTCODE 'N' >] -> Moved North
      | [< 'ACTCODE 'S' >] -> Moved South
      | [< 'ACTCODE 'E' >] -> Moved East
      | [< 'ACTCODE 'W' >] -> Moved West
      | [< 'ACTCODE 'P'; 'INT id >] -> Picked id
      | [< 'ACTCODE 'D'; 'INT id >] -> Dropped id
      | [< 'ACTCODE 'X'; 'INT x;
           'ACTCODE 'Y'; 'INT y; >] -> AddRobot(x,y)

    let rec action_list = parser
        [< a = action; rest = action_list >] -> a :: rest
      | [< >] -> []  (* don't eat the stop - action_report needs it *)

    let rec action_report = parser
        [< 'ID id; acts = action_list; rest = action_report >] ->
          (id,acts) :: rest
      | [< 'STOP >] -> []

    let rec package_list = parser
        [< 'INT id; 'INT x; 'INT y; 'INT weight;
           rest = package_list >] -> (id,x,y,weight) :: rest
      | [< 'STOP >] -> []


    (* Output a list of integers (for pick/drop lists). *)
    let rec output_int_list out_ch = function
        [] -> ()
      | n :: rest ->
          Printf.fprintf out_ch " %d" n;
          output_int_list out_ch rest

    (* Send a command to the server. *)
    let output_command out_ch bid command =
      Printf.fprintf out_ch "%d " bid;
      (match command with
        Move North -> output_string out_ch "Move N"
      | Move South -> output_string out_ch "Move S"
      | Move East -> output_string out_ch "Move E"
      | Move West -> output_string out_ch "Move W"
      | Pick list ->
          output_string out_ch "Pick";
          output_int_list out_ch list;
      | Drop list ->
          output_string out_ch "Drop";
          output_int_list out_ch list;);
      output_char out_ch '\n';
      flush out_ch

    (* Some test cases *)
    let tokens_of_string s = tokenize (Stream.of_string s) in begin
      assert ((Stream.next (tokens_of_string "#")) = TILE Wall);
      assert ((Stream.next (tokens_of_string "#5")) = ID 5);
      assert ((map (tokens_of_string "3 3\n..#\n#~@\n##.\n"))
                = {Map.size=(3,3); Map.tiles=[|[| Open; Open; Wall |];
                                               [| Wall; Water; HomeBase |];
                                               [| Wall; Wall; Open |]|]});
      assert ((config (tokens_of_string "10 50 1000\n")) = (10,50,1000));
      assert ((action_list [< 'ACTCODE 'X'; 'INT 1;
                              'ACTCODE 'Y'; 'INT 1; 'STOP >])
                = [AddRobot(1,1)]);
      assert ((action_list [< 'ACTCODE 'P'; 'INT 0;
                              'ACTCODE 'P'; 'INT 1; 'STOP >])
                = [Picked 0; Picked 1]);
      assert ((action_report (tokens_of_string "#1 E\n"))
                = [(1, [Moved East])]);
      assert ((action_report (tokens_of_string "#1 D 0 D 2 #5 N W S\n"))
                = [(1, [Dropped 0; Dropped 2]);
                   (5, [Moved North; Moved West; Moved South])]);
      assert ((package_list (tokens_of_string "1 2 3 4 5 6 7 8\n"))
                = [(1,2,3,4); (5,6,7,8)]);
      assert ((package_list (tokens_of_string "\n")) = []);
    end

    exception ConnectionClosed

    (* Create a class to encapsulate all that. *)
    open Unix
    class server name port =
      let host = gethostbyname name in
      let addr = ADDR_INET(host.h_addr_list.(0), port) in
      object (self)
        val mutable out_ch = Pervasives.stdout
        val mutable in_str = [< >]
        val mutable token_str = [< >]
        method connect = 
          let (sock_in, sock_out) = open_connection addr in
          out_ch <- sock_out;
          in_str <- Stream.of_channel sock_in;
          token_str <-  tokenize in_str;
          output_string out_ch "Player\n";
          flush out_ch;
        method private dump_end_message =
          Stream.iter (fun c -> print_char c) in_str;
          print_newline
        method recv_map = new Map.map (map token_str)
        method recv_config = config token_str
        method recv_action_report =
          try action_report token_str
          with _ -> (self#dump_end_message(); raise ConnectionClosed)
        method recv_package_list =
          try package_list token_str
          with _ -> (self#dump_end_message(); raise ConnectionClosed)
        method send_command = output_command out_ch
      end

  end;;


(* The PathFindingProblem and PathSearch are to find routes on the map. *)
module PathFindingProblem =
  struct
    type state = (int * int) * Map.map
    type action = direction
    let operators = [North; South; East; West]

    let apply (loc, m) move = ((next_location loc move), m)

    let action_valid (loc, m) move =
      let tile = m#get (next_location loc move) in
      tile = Open || tile = HomeBase

    let cost _ = 1

    let print_state ((x,y),_) move =
      Printf.printf "at %d %d" x y;
      (match move with
        Some d -> print_string " ... going ";
          print_char (char_of_direction d);
      | _ -> ());
      print_newline()

    let default_state = (0,0), Map.default_map
    class instance (init_state:state) goal_state =
      object
        method initialState = init_state
        method goalTest (state:state) = (state = goal_state)
        method evalState (((x,y),_):state) =
          let goal_x, goal_y = fst goal_state in
          (abs (goal_x - x)) + (abs (goal_y - y))
      end

    module StateCache =
      struct
        type t = (int*int, bool) Hashtbl.t
        let create () = Hashtbl.create 25
        let clear (cache:t) = Hashtbl.clear cache
        let add (cache:t) (loc,map) = Hashtbl.add cache loc true
        let mem (cache:t) (loc,map) = Hashtbl.mem cache loc
      end

  end;;

module PathSearch = Search.AStarSearch(PathFindingProblem);;

let find_route map here there =
  let pathProblem = new PathFindingProblem.instance (here,map) (there,map)
  in match PathSearch.search pathProblem with
    PathSearch.NoSolution -> None
  | PathSearch.Solution n ->
      let states = PathSearch.get_sequence n in
      let states' = List.map (fun (dir,(loc,map)) -> (dir,loc)) states in
      Some states';;


(* This module represents the robot agent. *)
module Agent =
  struct

    (* Types for knowledge about the world. *)
    module HomeBase = struct
      (* Home bases - keep track of whether they have been explored *)
      (* and where they are located. *)
      type homeBase = { mutable explored : bool; location : int*int }

      (* Find all the home bases on a map. *)
      (* Returns an ((int*int),homeBase) Hashtbl.t. *)
      let find_home_bases map =
        let bases = Hashtbl.create 8 in
        for i = 1 to map#width do
          for j = 1 to map#height do
            if map#get (i,j) = HomeBase
            then Hashtbl.add bases (i,j) { explored=false; location=(i,j) }
          done
        done;
        bases

      (* Update the base at this location to indicate it has been explored.*)
      let explored_base bases loc =
        try
          let base = Hashtbl.find bases loc in
          base.explored <- true
        with Not_found -> ()

      (* Get a list of all unexplored bases. *)
      let unexplored_bases bases =
        let unexplored = ref [] in
        Hashtbl.iter (fun _ base -> if not base.explored then
          unexplored := base :: !unexplored)
          bases;
        !unexplored
    end

    (* Robot - information about the other robot agents. *)
    module Robot = struct
      type robot = { id : int; mutable location : int * int }
      (* Robot set funcs. *)
      let create_set () = Hashtbl.create 13
      let find robots id =
        try
          Some (Hashtbl.find robots id)
        with Not_found -> None
      let add = Hashtbl.add
      let mem = Hashtbl.add

      (* Update a robot with an event. *)
      let event_move_robot robots id dir =
        match find robots id with
          Some robot ->
            let new_location = next_location robot.location dir in
            Printf.printf "robot #%d moves %c -> %d %d\n"
              id (char_of_direction dir) (fst new_location) (snd new_location);
            (robot.location <- new_location)
        | None -> ()
      let event_add_robot robots id loc =
        match find robots id with
          Some robot -> (robot.location <- loc)
        | None -> add robots id {id=id; location=loc}
    end

    (* Packages *)
    module Package = struct
      (* Status can be OnGround, Carried by some robot, or Unknown,    *)
      (* which means dropped at some location that may have been the   *)
      (* destination but we don't know. *)
      type packageStatus = Unknown | OnGround | Carried of int | Delivered

      type package = { id : int;
                       mutable status : packageStatus;
                       mutable location : int*int;
                       mutable seen : bool;
                       mutable destination : int * int;
                       mutable deliverable : bool;
                       mutable weight : int; }

      (* Package set funcs. *)
      let create_set () = Hashtbl.create 103
      let find packages id =
        try
          Some (Hashtbl.find packages id)
        with Not_found -> None
      let add = Hashtbl.add
      let mem = Hashtbl.add

      (* Update a package set with package details. *)
      let add_details packages location map details =
        let check_pkg (id,dest_x,dest_y,weight) =
          match find packages id with
            None -> add packages id { id=id;
                                      status=OnGround;
                                      seen=true;
                                      location=location;
                                      destination=(dest_x,dest_y);
                                      deliverable=(find_route map location (dest_x,dest_y) <> None);
                                      weight=weight; };
              Printf.printf "New package #%d, going to %d,%d, weighing %d\n"
                id dest_x dest_y weight;
          | Some pkg ->
              pkg.location <- location;
              if not pkg.seen then begin
                pkg.status <- OnGround;
                pkg.seen <- true;
                pkg.destination <- (dest_x,dest_y);
                pkg.deliverable <- find_route map location (dest_x,dest_y) <> None;
                pkg.weight <- weight;
                Printf.printf "Details for #%d, going to %d,%d, weighing %d\n"
                  id dest_x dest_y weight;
              end;
        in List.iter check_pkg details

      (* Handle a package event - pick up or drop. *)
      let event_pick packages pkgID robotID =
        match (find packages pkgID) with
          Some pkg -> (pkg.status <- Carried robotID)
        | None -> add packages pkgID { id=pkgID;
                                       status=Carried robotID;
                                       seen=false;
                                       location=(0,0);
                                       destination=(0,0);
                                       deliverable=false;
                                       weight=0; }
      let event_drop packages pkgID robots robotID =
        match (find packages pkgID, Robot.find robots robotID) with
          (Some pkg, Some robot) -> (pkg.status <-
            (if pkg.seen then begin
              Printf.printf "package for (%d,%d) dropped at (%d,%d)\n"
                (fst pkg.destination) (snd pkg.destination)
                (fst robot.Robot.location) (snd robot.Robot.location);
              (if robot.Robot.location = pkg.destination then Delivered
              else OnGround)
            end
            else Unknown))
        | (_,_) -> ()

      let iter packages f =
        let result = ref [] in
        let add_if id pkg = if f id pkg then result := pkg :: !result in
        Hashtbl.iter add_if packages;
        !result

      (* Get a list of packages that are available to be picked up. *)
      let available_for_pickup packages =
        iter packages (fun id pkg -> pkg.status = OnGround && pkg.deliverable)

      (* Get a list of packages that are carried by the robot. *)
      let carried_by_robot packages robotID =
        iter packages (fun id pkg -> pkg.status = Carried robotID)

      (* Get a list of packages at the given location. *)
      let at_location packages loc =
        iter packages (fun id p -> p.status = OnGround && p.location = loc && p.deliverable)

      (* Get a list of packages to be delivered to the given location. *)
      let for_location packages loc =
        iter packages (fun id pkg -> pkg.destination = loc)
    end


    (* Given a list of possible destinations, find the route to the closest. *)
    let find_route_to_closest map loc dests =
      let routes = List.map (function None -> [] | Some r -> r)
          (List.filter (fun r -> r <> None)
             (List.map (fun dest -> find_route map loc dest) dests)) in
      let route_lengths = List.map List.length routes in
      let route,_ = List.fold_left2 (fun (r,len) r2 len2 ->
        if len2 < len then (r2,len2) else (r,len))
          ([], 65535)  (* bogus starter value with big length *)
          routes route_lengths
      in
      (* Turn route (a list of directions) into a list of commands. *)
      List.map (fun (dir,loc) -> (Move dir,loc)) route

    (* Find the closest unexplored base and return the route. *)
    let explore_base bases map loc =
      let candidates = HomeBase.unexplored_bases bases in
      let locations = List.map (fun base -> base.HomeBase.location) candidates
      in find_route_to_closest map loc locations

    (* Find the closest package available for pickup and return the route. *)
    let route_to_pickup packages map loc =
      let candidates = Package.available_for_pickup packages in
      let locations = List.map (fun pkg -> pkg.Package.location) candidates in
      find_route_to_closest map loc locations

    (* Goals *)
    module Goal = struct
      type goal = Explore | Pickup | Deliver

      let pick_a_goal robotID location capacity packages =
        let carried = Package.carried_by_robot packages robotID in
        let available = Package.available_for_pickup packages in
        let packages_here = Package.at_location packages location in
        let sizes_here = List.map (fun p -> p.Package.weight) packages_here in
        if ((List.length packages_here > 0) &&
          (List.exists (fun w -> w < capacity) sizes_here))
        then Pickup
        else if List.length carried > 0 then Deliver
        else if List.length available > 0 then Pickup
        else Explore

      let actions_for_goal goal map robotID location capacity packages bases =
        match goal with
          Explore -> explore_base bases map location;
        | Pickup ->
            let route = route_to_pickup packages map location in
            if List.length route > 0 then route
            else let choices = List.filter
                (fun pkg -> pkg.Package.weight <= capacity)
                (Package.at_location packages location) in
            let best = List.sort
                (fun p1 p2 -> compare p2.Package.weight p1.Package.weight)
                choices in
            let (carrying, pickIDs) = List.fold_left
                (fun (total,pkgs) pkg ->
                  if total + pkg.Package.weight <= capacity then
                    (total + pkg.Package.weight, pkg.Package.id :: pkgs)
                  else (total,pkgs))
                (0,[]) best in
            Printf.printf "picking up %d of packages " carrying;
            List.iter (fun id -> Printf.printf " #%d" id) pickIDs;
            print_newline();
            [(Pick pickIDs, location)]
        | Deliver ->
            let carried = Package.carried_by_robot packages robotID in
            let dropPts = List.map (fun p -> p.Package.destination) carried in
            let route = find_route_to_closest map location dropPts in
            if List.length route > 0 then route
            else let dropoffs = List.filter
                (fun pkg -> pkg.Package.destination = location)
                carried in
            let dropIDs = List.map (fun p -> p.Package.id) dropoffs in
            print_string "delivering packages ";
            List.iter (fun id -> Printf.printf " #%d" id) dropIDs;
            print_newline();
            [(Drop dropIDs, location)]
    end


    (* The robot class defines the interface used by the main loop. *)
    class robot (id:int) capacity money map =
      object (self)
        val mutable money = money
        val mutable capacity = capacity
        val mutable location = (0,0)
        val bases = HomeBase.find_home_bases map
        val packages = Package.create_set()
        val robots = Robot.create_set()
        val mutable goal = Goal.Explore
        val mutable actions = []
        val mutable expectedLocation = (0,0)

        (* Handle update info for location of self. *)
        method private update_self events =
          let do_update = function
            | AddRobot loc ->
                location <- loc;
                expectedLocation <- loc;
                Printf.printf "setting initial location: %d %d\n"
                  (fst loc) (snd loc);
            | Moved dir -> (location <- next_location location dir);
            | Picked id ->
                (match Package.find packages id with
                  Some pkg -> (capacity <- capacity - pkg.Package.weight)
                | None -> ())
            | Dropped id ->
                (match Package.find packages id with
                  Some pkg -> (capacity <- capacity + pkg.Package.weight)
                | None -> ())
          in List.iter do_update events;
          (* any base here is now explored *)
          HomeBase.explored_base bases location;

        (* Update memory of other objects. *)
        method private update_others events =
          let do_update robotID = function
            | Picked pkgID -> Package.event_pick packages pkgID robotID
            | Dropped pkgID -> Package.event_drop packages pkgID robots robotID
            | Moved dir -> Robot.event_move_robot robots robotID dir
            | AddRobot loc -> Robot.event_add_robot robots robotID loc
          in
          List.iter (fun (robotID,bot_events) ->
            List.iter (do_update robotID) bot_events)
            events

        (* Check the goal - see if there's maybe something else
           we should be doing. *)
        method check_goal () =
          let new_goal = Goal.pick_a_goal id location capacity packages in
          (goal = new_goal)

        (* Check the plan - make sure we are where we wanted to be. *)
        method check_plan () = (location = expectedLocation)

        method robot_events events =
          self#update_self (List.assoc id events);
          self#update_others events;
        method package_details details =
          Package.add_details packages location map details
        method get_bid =
          let bid = 1 in
          money <- money - bid;
          bid
        method get_action =
          (* Make sure the goal is still valid. *)
          if not (self#check_goal()) then begin
            (* Abort old plan *)
            print_endline "changing plans";
            actions <- [];
            expectedLocation <- location;
          end;
          (* Make sure the plan is still valid for the goal. *)
          if not (self#check_plan()) then begin
            (* Abort old plan. *)
            print_endline "i'm off course - aborting";
            actions <- [];
          end;
          (* If we have no action plan stored, get one. *)
          if actions = [] then
            goal <- Goal.pick_a_goal id location capacity packages;
            actions <- Goal.actions_for_goal goal map id location capacity packages bases;
          (* Return the first command in the plan. *)
          if actions = [] then Pick [0]
          else let (command,expect) = List.hd actions in
          actions <- List.tl actions;
          expectedLocation <- expect;
          command
      end

  end;;


(* Test a server connection. *)
let map = ref Map.default_map;;
let turnCount = ref 0;;

let game server port =
  let server = new RDSServer.server server port in
  server#connect;

  map := server#recv_map;
  let (id, capacity, money) = server#recv_config in
  Printf.printf "robot #%d  capacity %d  money %d\n" id capacity money;

  let robot = new Agent.robot id capacity money !map in  
  try
    while true do begin
      incr turnCount;
      robot#robot_events server#recv_action_report;
      robot#package_details server#recv_package_list;
      server#send_command robot#get_bid robot#get_action;
    end done;
  with RDSServer.ConnectionClosed ->
    (* this is raised when the server closes the connection *)
    Printf.printf "took %d turns\n" !turnCount;
;;

let main () =
  let serverName = (if Array.length Sys.argv > 1 then
    Sys.argv.(1) else "localhost") in
  let serverPort = (if Array.length Sys.argv > 2 then
    int_of_string Sys.argv.(2) else 20005) in
  game serverName serverPort;;

main();;
