(* ICFP 2002 Programming Contest *)
(* Robot Delivery Service *)
(* Grant Glouser <glouser@cs.stanford.edu> *)

(* Generic framework for search problems. *)
(* Much of this is based on Ch 3 & 4 of "Artificial Intelligence,
   a Modern Approach" by Russell and Norvig. *)
(* Type of a generic problem. *)
module type PROBLEM =
  sig
    type state
    type action

    val operators : action list
    val apply : state -> action -> state
    val action_valid : state -> action -> bool
    val cost : action -> int

    val print_state : state -> action option -> unit

    class instance : state -> state ->
      object
        method initialState : state
        method goalTest : state -> bool
        method evalState : state -> int
      end

    (* Hash and comparison functions for state caching. *)
    module StateCache :
        sig
          type t
          val create : unit -> t
          val clear : t -> unit
          val add : t -> state -> unit
          val mem : t -> state -> bool
        end

  end;;


(* Priority queue - from the ocaml docs chapter 2 *)
module PrioQueue =
  struct
    type priority = int
    type 'a queue = Empty | Node of priority * 'a * 'a queue * 'a queue
    let empty = Empty
    let rec insert queue prio elt =
      match queue with
        Empty -> Node(prio, elt, Empty, Empty)
      | Node(p, e, left, right) ->
          if prio <= p
          then Node(prio, elt, insert right p e, left)
          else Node(p, e, insert right prio elt, left)
    exception Queue_is_empty
    let rec remove_top = function
        Empty -> raise Queue_is_empty
      | Node(prio, elt, left, Empty) -> left
      | Node(prio, elt, Empty, right) -> right
      | Node(prio, elt, (Node(lprio, lelt, _, _) as left),
             (Node(rprio, relt, _, _) as right)) ->
               if lprio <= rprio
               then Node(lprio, lelt, remove_top left, right)
               else Node(rprio, relt, left, remove_top right)
    let extract = function
        Empty -> raise Queue_is_empty
      | Node(prio, elt, _, _) as queue -> (prio, elt, remove_top queue)
   end;;


(* A* search algorithm. *)
module AStarSearch (Problem: PROBLEM) =
  struct

    (* The type of nodes in the search tree. *)
    type node = { state: Problem.state;
                  parent: node option;
                  operator: Problem.action option;
                  depth: int;
                  cost: int }

    (* The search tree - a priority queue for A* search. *)
    module Queue =
      struct
        let eval problem node = (problem#evalState node.state) + node.cost

        let make problem init_node = PrioQueue.insert PrioQueue.empty
            (eval problem init_node) init_node

        let empty = function PrioQueue.Empty -> true | _ -> false

        let remove_front queue =
          let (_, node, rest) = PrioQueue.extract queue in
          (node,rest)

        let queuing_fn problem =
          List.fold_left (fun q n -> PrioQueue.insert q (eval problem n) n)
      end

    (* Accumulators for statistics. *)
    let nodesExpanded = ref 0
    let childCount = ref 0
    let cacheEntries = ref 0
    let cacheHits = ref 0

    (* State cache. *)
    let cache = Problem.StateCache.create()

    (* Get the list of legal actions for a given state. *)
    let legalOperators node actions =
      List.filter (fun a -> (Problem.action_valid node.state a)) actions

    (* Get the list of successor states for a given state. *)
    let expand node actions =
      let legal_actions = legalOperators node actions in
      let new_nodes =
        List.map (fun a -> {state = (Problem.apply node.state a);
                            parent = (Some node);
                            operator = (Some a);
                            depth = node.depth + 1;
                            cost = node.cost + Problem.cost a })
          legal_actions
      in
      let not_in_cache = ref [] in
      begin
        (* Keep some statistics. *)
        if (List.length new_nodes) <> 0 then
          incr nodesExpanded;
        childCount := !childCount + (List.length new_nodes);

        (* Check each node against the cache, one by one.
           This way, we catch duplicates among the new nodes. *)
        List.iter (fun node ->
          if Problem.StateCache.mem cache node.state then
            (* Ignore this node - it's already in the cache. *)
            incr cacheHits
          else begin
            (* Add this node to the cache. *)
            Problem.StateCache.add cache node.state;
            incr cacheEntries;
            (* Add this node to the list of nodes to search. *)
            not_in_cache := node :: !not_in_cache;
          end)
          new_nodes;
        
        (* Return only the nodes that weren't in the cache. *)
        !not_in_cache;
      end

    (* The general search function. *)
    type searchResult = NoSolution | Solution of node
    exception StopSearch of searchResult

    let search (problem : Problem.instance) =
      (* nodes is the list of nodes to be searched *)
      let initialNode = { state = problem#initialState;
                          parent = None;
                          operator = None;
                          depth = 0;
                          cost = 0 } in
      let nodes = ref (Queue.make problem initialNode) in

      try
        (* Reset the search statistics and clear the cache. *)
        nodesExpanded := 0;
        childCount := 0;
        cacheEntries := 0;
        cacheHits := 0;
        Problem.StateCache.clear cache;

        (* Check each node in the list until one of them passes the
           goal test or we run out of nodes. *)
        while true do
          if Queue.empty !nodes then raise (StopSearch NoSolution)
          else
            let (node, rest) = Queue.remove_front !nodes in
            (* If this node passes the goal test, then return it. *)
            if (problem#goalTest node.state) then
              raise (StopSearch (Solution node))

            (* Otherwise, expand the node. *)
            else
              (* Add the new nodes to the queue. *)
              nodes := Queue.queuing_fn problem rest (expand node Problem.operators);
        done;
        NoSolution;  (* ... should never reach here ... *)

      (* StopSearch is raised when a result is found. *)
      with StopSearch result -> result


    (* Get a sequence of actions from a seach node. *)
    let get_action_sequence node =
      let rec build_seq seq = function
          {state=_; parent=Some p; operator=Some o; depth=_; cost=_} ->
            build_seq (o :: seq) p
        | _ -> seq in
      build_seq [] node

    (* Get a sequence of (action,state) from a seach node. *)
    (* That's the action and the state it should result it. *)
    let get_sequence node =
      let rec build_seq seq = function
          {state=s; parent=Some p; operator=Some o; depth=_; cost=_} ->
            build_seq ((o,s) :: seq) p
        | _ -> seq in
      build_seq [] node

    (* Generic output sequence printer. *)
    let rec print_sequence node op =
      match node with
        {state=s; parent=None; operator=_; depth=_; cost=_} ->
          Problem.print_state s op
      | {state=s; parent=(Some p); operator=o; depth=_; cost=_} ->
          begin
            print_sequence p o;
            Problem.print_state s op;
          end

  end;;
